from pathlib import Path
from tomllib import load
from numpy.lib.npyio import loadtxt
from science_signal import to_signal # type: ignore[reportMissingTypeStubs]
from science_signal.signal import Signal # type: ignore[reportMissingTypeStubs]
from scipy.io.matlab import loadmat # type: ignore[reportMissingTypeStubs]
from backscattering_analyzer import logger
from backscattering_analyzer.acquisition import (
    Acquisition,
    AcquisitionType,
)
from backscattering_analyzer.movements import Movements


class Data:
    def __init__(self, folder: Path, bench: str, date: str):
        logger.debug(
            _("loading data for {bench} in {date}").format(
                bench=bench, date=date
            )
        )
        folders: dict[str, Path] = {}  # list of folder
        files: dict[str, Path] = {}  # list of file

        folders["experiment"] = folder / "experiment_{date}".format(
            date=date,
        )
        files["metadata"] = folders["experiment"] / "metadata.toml"

        try:
            with open(files["metadata"], "rb") as file:
                metadata = load(file)
            logger.debug(
                _("successfully load {metadata}").format(
                    metadata=files["metadata"],
                )
            )
        except OSError as error:
            logger.critical(
                _("cannot load {metadata}: {error}").format(
                    metadata=files["metadata"],
                    error=error,
                )
            )
            raise error

        try:
            available_benches: list[str] = metadata["general"][
                "benches"
            ]
        except KeyError:
            logger.critical(
                _("malformed metadata, no benches provided")
            )
            raise Exception(
                _("malformed metadata, no benches provided")
            )

        if bench not in available_benches:
            logger.critical(
                _("no data on {bench} in {date}").format(
                    bench=bench,
                    data=date,
                )
            )
            raise Exception(
                _("no data on {bench} in {date}").format(
                    bench=bench,
                    data=date,
                )
            )

        description: str | None = None
        try:
            description = metadata["general"]["description"]
        except KeyError:
            logger.warning(
                _("malformed metadata, no description provided")
            )

        logger.info(
            "data description: {description}".format(
                description=description
            )
        )
        metadata_reference: dict[str, float]
        metadata_excited: dict[str, float]
        excited = False
        try:
            metadata_reference = metadata["benches"][bench]["reference"]
            folders["reference"] = folders[
                "experiment"
            ] / "{bench}/reference".format(
                bench=bench,
            )
            logger.debug(
                _("there are multiple reference for each benches")
            )
        except KeyError:
            try:
                metadata_reference = metadata["reference"]
                folders["reference"] = (
                    folders["experiment"] / "reference"
                )
                logger.debug(
                    _("there is one reference for every benches")
                )
            except KeyError:
                logger.critical(
                    _("malformed metadata, no reference provided")
                )
                raise Exception(
                    _("malformed metadata, no reference provided")
                )

        try:
            metadata_excited = metadata["benches"][bench]["excited"]
            folders["excited"] = folders[
                "experiment"
            ] / "{bench}/excited".format(
                bench=bench,
            )
            excited = True
            logger.debug(_("excited data have been taken"))
        except KeyError:
            logger.debug(_("no excited data have been taken"))

        logger.debug(_("metadata fully read, starting to load data"))

        channel_names = channel_name_mapping(bench)

        files["ref_bench"] = folders[
            "reference"
        ] / "{channel}.csv".format(channel=channel_names["bench"])
        files["ref_mirror"] = folders[
            "reference"
        ] / "{channel}.csv".format(channel=channel_names["mirror"])
        files["ref_sensibility"] = folders[
            "reference"
        ] / "{channel}.csv".format(channel=channel_names["sensibility"])

        time: float
        duration: float

        try:
            movements = Movements(
                bench=to_signal(loadtxt(files["ref_bench"]).T),
                mirror=to_signal(loadtxt(files["ref_mirror"]).T),
            )
            sensibility = to_signal(loadtxt(files["ref_sensibility"]).T)
            time = sensibility.x[0]
            duration = sensibility.x[-1] - sensibility.x[0]
            logger.debug(
                _(
                    "successfully loaded reference data files: "
                    + "\n {bench}"
                    + "\n {mirror}"
                    + "\n {sensibility}"
                ).format(
                    bench=files["ref_bench"],
                    mirror=files["ref_mirror"],
                    sensibility=files["ref_sensibility"],
                )
            )
            try:
                time = metadata_reference["time"]
                duration = metadata_reference["duration"]
                if sensibility.x[0] != time:
                    logger.warning(
                        _(
                            "time between metadata and data does not match: "
                            + "{data} != {metadata}"
                        ).format(
                            data=sensibility.x[0],
                            metadata=time,
                        )
                    )
                if abs(
                    sensibility.x[-1] - sensibility.x[0] - duration
                ) > 1e-2:
                    logger.warning(
                        _(
                            "duration between metadata and data does not "
                            + "match: {data} != {metadata}"
                        ).format(
                            data=(sensibility.x[-1] - sensibility.x[0]),
                            metadata=duration,
                        )
                    )
            except KeyError:
                logger.warning(
                    _(
                        "malformed metadata, each reference entry should have "
                        + "its own time and duration. Defining it with data"
                    )
                )
        except OSError as e:
            logger.critical(
                _(
                    "some files could not be read: {error}".format(
                        error=e
                    )
                )
            )
            raise e

        self.reference = Acquisition(
            time=time,
            duration=duration,
            movements=movements,
            sensibility=sensibility,
            type=AcquisitionType.REFERENCE,
        )

        logger.debug(_("reference data loaded and checked"))

        if excited:
            files["exc_bench"] = folders[
                "excited"
            ] / "{channel}.csv".format(channel=channel_names["bench"])
            files["exc_mirror"] = folders[
                "excited"
            ] / "{channel}.csv".format(channel=channel_names["mirror"])
            files["exc_sensibility"] = folders[
                "excited"
            ] / "{channel}.csv".format(
                channel=channel_names["sensibility"]
            )
            try:
                movements = Movements(
                    bench=to_signal(loadtxt(files["exc_bench"]).T),
                    mirror=to_signal(loadtxt(files["exc_mirror"]).T),
                )
                sensibility = to_signal(
                    loadtxt(files["exc_sensibility"]).T
                )
                time = sensibility.x[0]
                duration = sensibility.x[-1] - sensibility.x[0]
                logger.debug(
                    _(
                        "successfully loaded excited data files: "
                        + "\n {bench}"
                        + "\n {mirror}"
                        + "\n {sensibility}"
                    ).format(
                        bench=files["exc_bench"],
                        mirror=files["exc_mirror"],
                        sensibility=files["exc_sensibility"],
                    )
                )
                try:
                    time = metadata_excited["time"]
                    duration = metadata_excited["duration"]
                    if sensibility.x[0] != time:
                        logger.warning(
                            _(
                                "time between metadata and data does not "
                                + "match: {data} != {metadata}"
                            ).format(
                                data=sensibility.x[0],
                                metadata=time,
                            )
                        )
                    if abs(
                        sensibility.x[-1] - sensibility.x[0] - duration
                    ) > 1e-2:
                        logger.warning(
                            _(
                                "duration between metadata and data does "
                                + "not match: {data} != {metadata}"
                            ).format(
                                data=(
                                    sensibility.x[-1] - sensibility.x[0]
                                ),
                                metadata=duration,
                            )
                        )
                except KeyError:
                    logger.warning(
                        _(
                            "malformed metadata, each excited entry should "
                            + "have its own time and duration. Defining it "
                            + "with data"
                        )
                    )
            except OSError as e:
                logger.critical(
                    _(
                        "some files could not be read: {error}".format(
                            error=e
                        )
                    )
                )
            self.excited = Acquisition(
                time=time,
                duration=duration,
                movements=movements,
                sensibility=sensibility,
                type=AcquisitionType.EXCITED,
            )
            logger.debug(_("excited data loaded and checked"))


def channel_name_mapping(bench: str) -> dict[str, str]:
    names = {
        "sensibility": "h",
        "bench": bench,
    }
    if bench in ["SDB1", "SDB2"]:
        names["mirror"] = "SR"
    elif bench in ["SNEB", "SWEB"]:
        names["mirror"] = bench[1:3]
    elif bench == "SIB2":
        names["mirror"] = "MC"
    elif bench == "SPRB":
        names["mirror"] = "SR"  # not good, choice between MC, SR and PR
    else:
        raise ValueError(_("Unknow bench {bench}".format(bench=bench)))
    return names


def load_coupling(
    folder: Path,
    bench: str,
    date: str,
    file: str,
    powers: dict[str, float],
    model_powers: dict[str, float],
) -> list[Signal]:
    """
    Load and correct coupling date from modelisation
    """
    logger.debug(
        _("loading coupling for {bench} in {date}").format(
            bench=bench, date=date
        )
    )
    modelisation_file = folder / file
    try:
        modelisation_data = loadmat(modelisation_file)
        logger.debug(
            _("file successfully loaded:\n{file}").format(
                file=modelisation_file
            )
        )
    except OSError:
        logger.critical(
            _("cannot read {file}").format(file=modelisation_file)
        )
        raise Exception(
            _("cannot read {file}").format(file=modelisation_file)
        )
    if bench == "SDB2":
        modelisation_name = "SDB1coupling"
    elif bench == "SIB2":
        modelisation_name = "SIB1coupling"
    else:
        modelisation_name = "{bench}coupling".format(bench=bench)

    try:
        modelisation_signal = [
            Signal(
                modelisation_data["freq"][0],
                abs(coupling),
            )
            for coupling in modelisation_data[modelisation_name]
        ]
    except KeyError:
        logger.critical(
            _("{name} is not defined in the modelisation data").format(
                name=modelisation_name
            )
        )
        raise Exception(
            _("{name} is not defined in the modelisation data").format(
                name=modelisation_name
            )
        )

    if model_powers["laser"] != 0:
        if bench in ["SDB1", "SDB2"]:
            modelisation_signal[0] = (
                modelisation_signal[0]
                * (powers["laser"] / model_powers["laser"])
                * (powers["dark_fringe"] / model_powers["dark_fringe"])
                ** (1 / 2)
            )  # radiation
            modelisation_signal[1] = modelisation_signal[1] * (
                powers["dark_fringe"] / model_powers["dark_fringe"]
            ) ** (1 / 2)  # phase
        elif bench in ["SWEB", "SNEB"]:
            modelisation_signal[1] = (
                modelisation_signal[1]
                * powers["laser"]
                / model_powers["laser"]
            )  # radiation
        elif bench in ["SIB2", "SPRB"]:
            logger.warning(
                _(
                    "cannot fix projection power for SIB2 and SPRB coupling"
                )
            )
    return modelisation_signal
